@@ -13,6 +13,7 @@ Parse and translate XML Schema regular expressions to Python regex syntax.
 import re
 from sys import maxunicode
 
+from ..helpers import OCCURRENCE_INDICATORS
 from .unicode_subsets import RegexError, UnicodeSubset, unicode_subset
 from .character_classes import I_SHORTCUT_REPLACE, C_SHORTCUT_REPLACE, CharacterClass
 
@@ -21,15 +22,16 @@ INVALID_HYPHEN_PATTERN = re.compile(r'[^\\]-[^\\]-[^\\]')
 DIGITS_PATTERN = re.compile(r'\d+')
 QUANTIFIER_PATTERN = re.compile(r'{\d+(,(\d+)?)?}')
 FORBIDDEN_ESCAPES_NOREF_PATTERN = re.compile(
-    r'(?<!\\)\\(U[0-9a-fA-F]{8}|u[0-9a-fA-F]{4}|x[0-9a-fA-F]{2}|o{\d+}|\d+|A|Z|z|B|b|o)'
+    r'(?<!\\)\\(U[0-9a-fA-F]{8}|u[0-9a-fA-F]{4}|x[0-9a-fA-F]{2}|o{\d+}|\d+|A|Z|z|B|b|o|0[0-9]{2})'
 )
 FORBIDDEN_ESCAPES_REF_PATTERN = re.compile(
-    r'(?<!\\)\\(U[0-9a-fA-F]{8}|u[0-9a-fA-F]{4}|x[0-9a-fA-F]{2}|o{\d+}|A|Z|z|B|b|o)'
+    r'(?<!\\)\\(U[0-9a-fA-F]{8}|u[0-9a-fA-F]{4}|x[0-9a-fA-F]{2}|o{\d+}|A|Z|z|B|b|o|0[0-9]{2})'
 )
 
 
-def translate_pattern(pattern: str, flags=0, xsd_version='1.0', back_references=True,
-                      lazy_quantifiers=True, anchors=True):
+def translate_pattern(pattern: str, flags: int = 0, xsd_version: str = '1.0',
+                      back_references: bool = True, lazy_quantifiers: bool = True,
+                      anchors: bool = True) -> str:
     """
     Translates a pattern regex expression to a Python regex pattern. With default
     options the translator processes XPath 2.0/XQuery 1.0 regex patterns. For XML
@@ -43,7 +45,10 @@ def translate_pattern(pattern: str, flags=0, xsd_version='1.0', back_references=
     :param anchors: if `True` supports ^ and $ anchors, otherwise the translated \
     pattern is anchored to its boundaries and anchors are treated as normal characters.
     """
-    def parse_character_class():
+    pos: int
+    msg: str
+
+    def parse_character_class() -> CharacterClass:
         nonlocal pos
         nonlocal msg
 
@@ -148,6 +153,11 @@ def translate_pattern(pattern: str, flags=0, xsd_version='1.0', back_references=
                 msg = "invalid quantifier {!r} at position {}: {!r}"
                 raise RegexError(msg.format(ch, pos, pattern))
 
+            if regex and regex[-1] in ('^', r'(?<!\n\Z)^', '$', r'$(?!\n\Z)'):
+                # ^{n} or ${n} allowed but useless. Invalid im Python re
+                # so incapsulate '^'/'$' inside a non-capturing group.
+                regex[-1] = f'(?:{regex[-1]})'
+
             regex.append(match.group())
             pos += len(match.group())
             if not lazy_quantifiers and pos < pattern_len and pattern[pos] in ('?', '+', '*'):
@@ -156,8 +166,8 @@ def translate_pattern(pattern: str, flags=0, xsd_version='1.0', back_references=
             continue  # pragma: no cover
 
         elif ch == '(':
-            if pattern[pos:pos + 2] == '(?':
-                msg = "invalid '(?...)' extension notation ad position {}: {!r}"
+            if pattern[pos:pos + 2] == '(?' and pattern[pos:pos + 3] != '(?:':
+                msg = "invalid '(?...)' extension notation at position {}: {!r}"
                 raise RegexError(msg.format(pos, pattern))
 
             total_groups += 1
@@ -176,7 +186,7 @@ def translate_pattern(pattern: str, flags=0, xsd_version='1.0', back_references=
             nested_groups -= 1
             regex.append(ch)
 
-        elif ch in ('?', '+', '*'):
+        elif ch in OCCURRENCE_INDICATORS:
             if pos == 0:
                 msg = "unexpected quantifier {!r} at position {}: {!r}"
                 raise RegexError(msg.format(ch, pos, pattern))
@@ -186,6 +196,11 @@ def translate_pattern(pattern: str, flags=0, xsd_version='1.0', back_references=
                 msg = "unexpected meta character {!r} at position {}: {!r}"
                 raise RegexError(msg.format(pattern[pos + 1], pos + 1, pattern))
 
+            if regex and regex[-1] in ('^', r'(?<!\n\Z)^', '$', r'$(?!\n\Z)'):
+                # ^*/^+/^? or $*/$+/$? allowed but useless. Invalid im Python re
+                # so incapsulate '^'/'$' inside a non-capturing group.
+                regex[-1] = f'(?:{regex[-1]})'
+
             regex.append(ch)
 
         elif ch == '\\':
